@@ -20,15 +20,31 @@
  *
  */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include "gmime-message.h"
 #include "gmime-utils.h"
 #include "internet-address.h"
-#include <config.h>
 #include <string.h>
 #include <ctype.h>
 #include <locale.h>
 
-#define GMIME_FOLD_LEN  76
+
+static char *rfc822_headers[] = {
+	"Return-Path",
+	"Received",
+	"Date",
+	"From",
+	"Reply-To",
+	"Subject",
+	"Sender",
+	"To",
+	"Cc",
+	NULL
+};
+
 
 /**
  * g_mime_message_new: Create a new MIME Message object
@@ -40,6 +56,8 @@ GMimeMessage *
 g_mime_message_new ()
 {
 	GMimeMessage *message;
+	GMimeHeader *headers;
+	int i;
 	
 	message = g_new0 (GMimeMessage, 1);
 	
@@ -47,8 +65,12 @@ g_mime_message_new ()
 	
 	message->header->recipients = g_hash_table_new (g_str_hash, g_str_equal);
 	
-	message->header->arbitrary_headers = g_ptr_array_new ();
+	message->header->headers = headers = g_mime_header_new ();
 	
+	/* Populate with the "standard" rfc822 headers so we can have a standard order */
+	for (i = 0; rfc822_headers[i]; i++) 
+		g_mime_header_set (headers, rfc822_headers[i], NULL);
+			   
 	return message;
 }
 
@@ -96,15 +118,9 @@ g_mime_message_destroy (GMimeMessage *message)
 	
 	g_free (message->header->subject);
 	
-	/* destroy arbitrary headers */
-	for (i = 0; i < message->header->arbitrary_headers->len; i++) {
-		GMimeHeader *header = message->header->arbitrary_headers->pdata[i];
-		
-		g_free (header->name);
-		g_free (header->value);
-		g_free (header);
-	}
-	g_ptr_array_free (message->header->arbitrary_headers, TRUE);
+	g_free (message->header->message_id);
+	
+	g_mime_header_destroy (message->header->headers);
 	
 	g_free (message->header);
 	
@@ -202,7 +218,7 @@ g_mime_message_add_recipient (GMimeMessage *message, gchar *type, const gchar *n
 	InternetAddress *ia;
 	GList *recipients;
 	
-	ia = internet_address_new (name, address);
+	ia = internet_address_new_name (name, address);
 	
 	recipients = g_hash_table_lookup (message->header->recipients, type);
 	g_hash_table_remove (message->header->recipients, type);
@@ -227,9 +243,7 @@ g_mime_message_add_recipient (GMimeMessage *message, gchar *type, const gchar *n
 void
 g_mime_message_add_recipients_from_string (GMimeMessage *message, gchar *type, const gchar *string)
 {
-	InternetAddress *ia;
-	GList *recipients;
-	gchar *ptr, *eptr, *recipient;
+	GList *recipients, *addrlist;
 	
 	g_return_if_fail (message != NULL);
 	g_return_if_fail (string != NULL);
@@ -237,38 +251,9 @@ g_mime_message_add_recipients_from_string (GMimeMessage *message, gchar *type, c
 	recipients = g_hash_table_lookup (message->header->recipients, type);
 	g_hash_table_remove (message->header->recipients, type);
 	
-	ptr = (gchar *) string;
-	
-	while (*ptr) {
-		gboolean in_quotes = FALSE;
-		
-		/* skip through leading whitespace */
-		for ( ; *ptr && isspace (*ptr); ptr++);
-		
-		if (*ptr == '"')
-			in_quotes = TRUE;
-		
-		/* find the end of this address */
-		eptr = ptr + 1;
-		while (*eptr) {
-			if (*eptr == '"' && *(eptr - 1) != '\\')
-				in_quotes = !in_quotes;
-			else if (*eptr == ',' && !in_quotes)
-				break;
-			
-			eptr++;
-		}
-		
-		recipient = g_strndup (ptr, (gint) (eptr - ptr));
-		ia = internet_address_new_from_string (recipient);
-		g_free (recipient);
-		recipients = g_list_append (recipients, ia);
-		
-		if (*eptr)
-			ptr = eptr + 1;
-		else
-			break;
-	}
+	addrlist = internet_address_parse_string (string);
+	if (addrlist)
+		recipients = g_list_concat (recipients, addrlist);
 	
 	g_hash_table_insert (message->header->recipients, type, recipients);
 }
@@ -393,7 +378,40 @@ g_mime_message_get_date_string (GMimeMessage *message)
 
 
 /**
- * g_mime_message_add_arbitrary_header: Add an arbitrary message header
+ * g_mime_message_set_message_id: 
+ * @message: MIME Message
+ * @id: message-id
+ *
+ * Set the Message-Id on a message.
+ **/
+void
+g_mime_message_set_message_id (GMimeMessage *message, const gchar *id)
+{
+	g_return_if_fail (message != NULL);
+	
+	if (message->header->message_id)
+		g_free (message->header->message_id);
+	message->header->message_id = g_strdup (id);
+}
+
+
+/**
+ * g_mime_message_get_message_id: 
+ * @message: MIME Message
+ *
+ * Returns the Message-Id of a message.
+ **/
+const gchar *
+g_mime_message_get_message_id (GMimeMessage *message)
+{
+	g_return_val_if_fail (message != NULL, NULL);
+	
+	return message->header->message_id;
+}
+
+
+/**
+ * g_mime_message_set_header: Add an arbitrary message header
  * @message: MIME Message
  * @field: rfc822 header field
  * @value: the contents of the header field
@@ -402,17 +420,27 @@ g_mime_message_get_date_string (GMimeMessage *message)
  * X-Priority, or In-Reply-To.
  **/
 void
-g_mime_message_add_arbitrary_header (GMimeMessage *message, const gchar *field, const gchar *value)
+g_mime_message_set_header (GMimeMessage *message, const gchar *field, const gchar *value)
 {
-	GMimeHeader *header;
-	
 	g_return_if_fail (message != NULL);
 	
-	header = g_new (GMimeHeader, 1);
-	header->name = g_strdup (field);
-	header->value = g_strdup (value);
+	g_mime_header_set (message->header->headers, field, value);
+}
+
+
+/**
+ * g_mime_message_get_header:
+ * @message: MIME Message
+ * @field: rfc822 header field
+ *
+ * Returns the value of the requested header (or NULL if it isn't set)
+ **/
+const gchar *
+g_mime_message_get_header (GMimeMessage *message, const gchar *field)
+{
+	g_return_val_if_fail (message != NULL, NULL);
 	
-	g_ptr_array_add (message->header->arbitrary_headers, header);
+	return g_mime_header_get (message->header->headers, field);
 }
 
 
@@ -434,137 +462,35 @@ g_mime_message_set_mime_part (GMimeMessage *message, GMimePart *mime_part)
 	message->mime_part = mime_part;
 }
 
-static gchar *
-header_fold (const gchar *in)
-{
-	const gchar *inptr, *space;
-	gint len, outlen, i;
-	GString *out;
-	gchar *ret;
-	
-	inptr = in;
-	len = strlen (in);
-	if (len <= GMIME_FOLD_LEN)
-		return g_strdup (in);
-	
-	out = g_string_new ("");
-	outlen = 0;
-	while (*inptr) {
-		space = strchr (inptr, ' ');
-		if (space)
-			len = space - inptr + 1;
-		else
-			len = strlen (inptr);
-		
-		if (outlen + len > GMIME_FOLD_LEN) {
-			g_string_append (out, "\n\t");
-			outlen = 1;
-			
-			/* check for very long words, just cut them up */
-			while (outlen + len > GMIME_FOLD_LEN) {
-				for (i = 0; i < GMIME_FOLD_LEN - outlen; i++)
-					g_string_append_c (out, inptr[i]);
-				inptr += GMIME_FOLD_LEN - outlen;
-				len -= GMIME_FOLD_LEN - outlen;
-				g_string_append (out, "\n\t");
-				outlen = 1;
-			}
-		}
-		
-		outlen += len;
-		for (i = 0; i < len; i++) {
-			g_string_append_c (out, inptr[i]);
-		}
-		inptr += len;
-	}
-	
-	ret = out->str;
-	g_string_free (out, FALSE);
-	
-	return ret;	
-}
 
-
-/**
- * g_mime_header_printf: Format a header.
- * @format: string format
- * @Varargs: arguments
- *
- * Returns an allocated string containing the folded header specified
- * by %format and the following arguments.
- **/
-gchar *
-g_mime_header_printf (const gchar *format, ...)
+static void
+sync_headers (GMimeMessage *message)
 {
-	gchar *buf, *ret;
-	va_list ap;
-	
-	va_start (ap, format);
-	buf = g_strdup_vprintf (format, ap);
-	va_end (ap);
-	
-	ret = header_fold (buf);
-	g_free (buf);
-	
-	return ret;
-}
-
-static gchar *
-create_header (GMimeMessage *message)
-{
-	/* Make sure to not append "\n\n" to the end
-	 * so we can add MIME headers later */
 	GString *string;
-	gchar *str, *buf, *date, *subject;
+	gchar *str, *date;
 	GList *recipients;
 	
-	string = g_string_new ("");
-	
-	/* write out the arbitrary headers first (as they may contain
-	 * "Received:" headers which really should come first) */
-	if (message->header->arbitrary_headers->len) {
-		gint i;
-		
-		for (i = 0; i < message->header->arbitrary_headers->len; i++) {
-			const GMimeHeader *header;
-			gchar *encoded_value;
-			
-			header = message->header->arbitrary_headers->pdata[i];
-			encoded_value = g_mime_utils_8bit_header_encode (header->value);
-			
-			buf = g_mime_header_printf ("%s: %s\n", header->name, encoded_value);
-			g_free (encoded_value);
-			
-			g_string_append (string, buf);
-			g_free (buf);
-		}
-	}
-	
-	/* create the standard headers */
+	/* sync the Date header */
 	if (!message->header->date)
 		g_mime_message_set_date (message, time (NULL), 0);
 	date = g_mime_message_get_date_string (message);
-	buf = g_mime_header_printf ("Date: %s\n", date);
-	g_string_append (string, buf);
+	g_mime_header_set (message->header->headers, "Date", date);
 	g_free (date);
-	g_free (buf);
 	
-	buf = g_mime_header_printf ("From: %s\n", message->header->from ? message->header->from : "");
-	g_string_append (string, buf);
-	g_free (buf);
+	/* sync the From header */
+	g_mime_header_set (message->header->headers, "From",
+			   message->header->from ? message->header->from : "");
 	
-	if (message->header->reply_to) {
-		buf = g_mime_header_printf ("Reply-To: %s\n", message->header->reply_to);
-		g_string_append (string, buf);
-		g_free (buf);
-	}
+	/* sync the Reply-To header */
+	g_mime_header_set (message->header->headers, "Reply-To", message->header->reply_to);
 	
+	/* sync the To header */
 	recipients = g_mime_message_get_recipients (message, GMIME_RECIPIENT_TYPE_TO);
 	if (recipients) {
 		GString *recip;
 		
-		recip = g_string_new ("To: ");
-		while (TRUE) {
+		recip = g_string_new ("");
+		while (recipients) {
 			InternetAddress *ia;
 			gchar *address;
 			
@@ -576,23 +502,19 @@ create_header (GMimeMessage *message)
 			recipients = recipients->next;
 			if (recipients)
 				g_string_append (recip, ", ");
-			else
-				break;
 		}
-		g_string_append (recip, "\n");
-		buf = header_fold (recip->str);
-		g_string_free (recip, TRUE);
 		
-		g_string_append (string, buf);
-		g_free (buf);
+		g_mime_header_set (message->header->headers, "To", recip->str);
+		g_string_free (recip, TRUE);
 	}
 	
+	/* sync the Cc header */
 	recipients = g_mime_message_get_recipients (message, GMIME_RECIPIENT_TYPE_CC);
 	if (recipients) {
 		GString *recip;
 		
-		recip = g_string_new ("Cc: ");
-		while (TRUE) {
+		recip = g_string_new ("");
+		while (recipients) {
 			InternetAddress *ia;
 			gchar *address;
 			
@@ -604,27 +526,40 @@ create_header (GMimeMessage *message)
 			recipients = recipients->next;
 			if (recipients)
 				g_string_append (recip, ", ");
-			else
-				break;
 		}
-		g_string_append (recip, "\n");
-		buf = header_fold (recip->str);
-		g_string_free (recip, TRUE);
 		
-		g_string_append (string, buf);
-		g_free (buf);
+		g_mime_header_set (message->header->headers, "Cc", recip->str);
+		g_string_free (recip, TRUE);
 	}
 	
-	subject = g_mime_utils_8bit_header_encode (message->header->subject);
-	buf = g_mime_header_printf ("Subject: %s\n", subject ? subject : "");
-	g_string_append (string, buf);
-	g_free (subject);
-	g_free (buf);
+	/* sync the Subject header */
+	g_mime_header_set (message->header->headers, "Subject", message->header->subject);
 	
-	str = string->str;
-	g_string_free (string, FALSE);
+	/* sync the Message-Id header */
+	g_mime_header_set (message->header->headers, "Message-Id", message->header->message_id);
+}
+
+
+/**
+ * g_mime_message_write_to_string: Write the MIME Message to a string
+ * @message: MIME Message
+ * @string: output string
+ *
+ * Write the contents of the MIME Message to @string.
+ **/
+void
+g_mime_message_write_to_string (GMimeMessage *message, GString *string)
+{
+	g_return_if_fail (message != NULL);
+	g_return_if_fail (string != NULL);
 	
-	return str;
+	sync_headers (message);
+	g_mime_header_write_to_string (message->header->headers, string);
+	
+	if (message->mime_part)
+		g_mime_part_write_to_string (message->mime_part, TRUE, string);
+	else
+		g_string_append (string, "\n");
 }
 
 
@@ -638,63 +573,56 @@ gchar *
 g_mime_message_to_string (GMimeMessage *message)
 {
 	GString *string;
-	gchar *str, *header, *body;
+	gchar *str;
 	
 	g_return_val_if_fail (message != NULL, NULL);
 	
-	header = create_header (message);
-	if (!header)
-		return NULL;
-	
-	string = g_string_new (header);
-	g_free (header);
-	
-	body = g_mime_part_to_string (message->mime_part, TRUE);
-	if (body)
-		g_string_append (string, body);
-	else
-		g_string_append (string, "\n");
-	g_free (body);
-	
+	string = g_string_new ("");
+	g_mime_message_write_to_string (message, string);
 	str = string->str;
 	g_string_free (string, FALSE);
 	
 	return str;
 }
 
-static gchar *
-decode_body_part (GMimePart *mime_part)
-{
-	GMimePartEncodingType encoding;
-	const gchar *content;
-	gchar *body;
-	int state = 0, save = 0;
-	
-	encoding = g_mime_part_get_encoding (mime_part);
-	content = mime_part->content;
-	
-	switch (encoding) {
-	case GMIME_PART_ENCODING_BASE64:
-		body = g_malloc0 (strlen (content));
-		g_mime_utils_base64_decode_step (content, strlen (content), body, &state, &save);
-		break;
-	case GMIME_PART_ENCODING_QUOTEDPRINTABLE:
-		body = g_malloc0 (strlen (content));
-		g_mime_utils_quoted_decode_step (content, strlen (content), body, &state, &save);
-		break;
-	default:
-		body = g_strdup (content);
-	}
-	
-	return body;
-}
 
+/* Brief explanation of how this function works it's magic:
+ *
+ * We cycle through the immediate subparts looking for text parts. If
+ * the first text part we come accross is exactly what we want then we
+ * return it, otherwise keep a reference to it for later use (if we
+ * don't find the preferred part later as we continue to cycle through
+ * the subparts then we default to the first text part found). If we
+ * come to a multipart, we descend into it repeating the process. If
+ * we find the 'body' in a sub-multipart, we don't necessarily return
+ * that value for it is entirely possible that there could be text
+ * parts defined after the sub-multipart. For example, we could have
+ * the following MIME structure:
+ *
+ * multipart/alternative
+ *   image/png
+ *   multipart/related
+ *     text/html
+ *     image/png
+ *     image/gif
+ *     image/jpeg
+ *   text/plain
+ *   text/html
+ *
+ * While one can never be certain that the text/html part within the
+ * multipart/related isn't the true 'body', it's genrally safe to
+ * assume that in cases like this, the outer text part(s) are the
+ * message body. Note that this is an assumption and is thus not
+ * guarenteed to always be correct.
+ **/
 static gchar *
-handle_multipart_alternative (GMimePart *multipart, gboolean want_plain, gboolean *is_html)
+multipart_get_body (GMimePart *multipart, gboolean want_plain, gboolean *is_html)
 {
-	GMimePart *last = NULL;
-	GList *child;
+	GMimePart *first = NULL;
+	const gchar *content;
 	gchar *body = NULL;
+	GList *child;
+	guint len;
 	
 	child = multipart->children;
 	while (child) {
@@ -705,27 +633,42 @@ handle_multipart_alternative (GMimePart *multipart, gboolean want_plain, gboolea
 		type = g_mime_part_get_content_type (mime_part);
 		
 		if (g_mime_content_type_is_type (type, "text", want_plain ? "plain" : "html")) {
-			/* we got what we wanted */
+			/* we got what we came for */
 			*is_html = !want_plain;
 			
-			body = decode_body_part (mime_part);
+			content = g_mime_part_get_content (mime_part, &len);
+			g_free (body);
+			body = g_strndup (content, len);
 			break;
-		} else {
-			last = mime_part;
+		} else if (g_mime_content_type_is_type (type, "text", "*") && !first) {
+			/* remember what our first text part was */
+			first = mime_part;
+			g_free (body);
+			body = NULL;
+		} else if (g_mime_content_type_is_type (type, "multipart", "*") && !first && !body) {
+			/* look in the multipart for the body */
+			body = multipart_get_body (mime_part, want_plain, is_html);
+			
+			/* You are probably asking: "why don't we break here?"
+			 * The answer is because the real message body could
+			 * be a part after this multipart */
 		}
 		
 		child = child->next;
 	}
 	
-	if (!body && last) {
+	if (!body && first) {
 		/* we didn't get the type we wanted but still got the body */
 		*is_html = want_plain;
-		body = decode_body_part (last);
+		
+		content = g_mime_part_get_content (first, &len);
+		body = g_strndup (content, len);
 	}
 	
 	return body;
 }
 
+
 /**
  * g_mime_message_get_body: Return the body of the message
  * @message: MIME Message
@@ -743,38 +686,23 @@ gchar *
 g_mime_message_get_body (const GMimeMessage *message, gboolean want_plain, gboolean *is_html)
 {
 	const GMimeContentType *type;
+	const gchar *content;
 	gchar *body = NULL;
+	guint len = 0;
 	
 	type = g_mime_part_get_content_type (message->mime_part);
 	if (g_mime_content_type_is_type (type, "text", "*")) {
+		/* this *has* to be the message body */
 		if (g_mime_content_type_is_type (type, "text", want_plain ? "plain" : "html"))
 			*is_html = !want_plain;
 		else
 			*is_html = want_plain;
 		
-		body = decode_body_part (message->mime_part);
-	} else if (g_mime_content_type_is_type (type, "multipart", "alternative")) {
-		/* Get the prefered part from the multipart/alternative */
-		body = handle_multipart_alternative (message->mime_part, want_plain, is_html);
+		content = g_mime_part_get_content (message->mime_part, &len);
+		body = g_strndup (content, len);
 	} else if (g_mime_content_type_is_type (type, "multipart", "*")) {
-		/* The message body has to be in the first MIME part of the multipart */
-		GMimePart *first_part;
-		
-		first_part = message->mime_part->children->data;
-		type = g_mime_part_get_content_type (first_part);
-		if (g_mime_content_type_is_type (type, "text", "*")) {
-			if (g_mime_content_type_is_type (type, "text", want_plain ? "plain" : "html"))
-				*is_html = !want_plain;
-			else
-				*is_html = want_plain;
-			
-			body = decode_body_part (first_part);
-		} else if (g_mime_content_type_is_type (type, "multipart", "alternative")) {
-			/* Get the prefered part from the multipart/alternative */
-			body = handle_multipart_alternative (first_part, want_plain, is_html);
-		} else {
-			/* fuck it, if we haven't found it yet we ain't gonna find it */
-		}
+		/* lets see if we can find a body in the multipart */
+		body = multipart_get_body (message->mime_part, want_plain, is_html);
 	}
 	
 	return body;
@@ -790,5 +718,26 @@ g_mime_message_get_body (const GMimeMessage *message, gboolean want_plain, gbool
 gchar *
 g_mime_message_get_headers (GMimeMessage *message)
 {
-	return create_header (message);
+	g_return_val_if_fail (message != NULL, NULL);
+	
+	sync_headers (message);
+	return g_mime_header_to_string (message->header->headers);
+}
+
+
+/**
+ * g_mime_message_foreach_part:
+ * @message: MIME message
+ * @callback: function to call on each of the mime parts contained by the mime message
+ * @data: extra data to pass to the callback
+ *
+ * Calls #callback on each of the mime parts in the mime message.
+ **/
+void
+g_mime_message_foreach_part (GMimeMessage *message, GMimePartFunc callback, gpointer data)
+{
+	g_return_if_fail (message != NULL);
+	g_return_if_fail (callback != NULL);
+	
+	g_mime_part_foreach (message->mime_part, callback, data);
 }
